import unittest
from argparse import ArgumentParser

from dnry.config.arg import ArgumentSource
from dnry.config import ConfigFactory


class TestArgSource(unittest.TestCase):
    def test_single_key(self):
        ap = ArgumentParser()
        ap.add_argument('-f', '--foo')
        fact = ConfigFactory()
        fact.add_source(ArgumentSource(ap, ["--foo", "bar"]))
        conf = fact.build()
        self.assertEqual("bar", conf.get("foo"))

    def test_single_key(self):
        ap = ArgumentParser()
        ap.add_argument('-f', '--foo')
        fact = ConfigFactory()
        fact.add_source(ArgumentSource(ap, ["--foo", "baz"]))
        conf = fact.build()
        self.assertEqual("baz", conf.get("foo"))

    def test_nested_key(self):
        ap = ArgumentParser()
        ap.add_argument('-f', '--foo:bar')
        fact = ConfigFactory()
        fact.add_source(ArgumentSource(ap, ["--foo:bar", "baz"]))
        conf = fact.build()
        self.assertEqual("baz", conf.get("foo:bar"))

    def test_many_nested_keys(self):
        ap = ArgumentParser()
        ap.add_argument('--foo:bar')
        ap.add_argument('--bar')
        ap.add_argument('--foo:baz')
        fact = ConfigFactory()
        fact.add_source(ArgumentSource(ap, ["--foo:bar", "a", "--bar", "b", "--foo:baz", "c"]))
        conf = fact.build()
        self.assertEqual("a", conf.get("foo:bar"))
        self.assertEqual("b", conf.get("bar"))
        self.assertEqual("c", conf.get("foo:baz"))



import os

import yaml
from yaml.parser import ParserError
import tempfile
import unittest

from dnry.config import ConfigFactory
from dnry.config.yaml import YamlSource


def make_config_with_temp(data: dict):
    _, file = tempfile.mkstemp()
    try:
        with open(file, 'w') as fd:
            yaml.dump(data, stream=fd, Dumper=yaml.SafeDumper)
        factory = ConfigFactory()
        factory.add_source(YamlSource(file))
        return factory.build()
    finally:
        os.unlink(file)


class TestYamlSource(unittest.TestCase):

    def test_single_key(self):
        conf = make_config_with_temp({"a": 1})
        val = conf.get("a")
        self.assertEqual(1, val)

    def test_deep_keys(self):
        conf = make_config_with_temp({"a": 1, "b": {"c": 2}})
        val = conf.get("b:c")
        self.assertEqual(2, val)

    def test_bad_yaml(self):
        _, file = tempfile.mkstemp()
        throw_error = False
        try:
            with open(file, 'w') as fd:
                fd.write(':')
            factory = ConfigFactory()
            factory.add_source(YamlSource(file))
            conf = factory.build()
        except ParserError:
            throw_error = True
        finally:
            os.unlink(file)

        self.assertTrue(throw_error)




from typing import List, Any

from dnry.config.section import ConfigSection
from dnry.config.helpers import merge
from dnry.config.types import IConfigFactory, IConfigSource, IConfigSection


class ConfigFactory(IConfigFactory):
    """Builds a single configuration section from a set of configuration sources.

    Configuration sources will be loaded and merged in order. If configuration sources
    collide key names, the value for the last source loaded containing the key will
    be used.

    You can override the merge behavior by providing your own implementation and setting it
    on ConfigurationFactory.merge.

    You can provide your own IConfigurationSection implementation by setting it on
    ConfigurationFactory.Section. Note: you must accept a single dict in __init__.
    """

    __explicit_configs: List[IConfigSection]
    __sources: List[IConfigSource]
    Section = ConfigSection
    merge = merge

    def __init__(self, sources: List[IConfigSource] = None):
        self.__explicit_configs = list()
        self.__sources = sources or list()

    def add_source(self, source: IConfigSource) -> None:
        self.__sources.append(source)

    def add_configuration(self, conf: IConfigSection):
        self.__explicit_configs.append(conf)

    def build(self) -> IConfigSection:
        context = {}
        for source in self.__sources:
            context = ConfigFactory.merge(context, source.load(self, ConfigFactory.Section(context)))
        for conf in self.__explicit_configs:
            context = ConfigFactory.merge(context, vars(conf))
        return ConfigFactory.Section(context)



from dnry.config.types import IConfigSource, IConfigFactory, IConfigSection


class InMemorySource(IConfigSource):
    """Read configuration values from a dictionary"""
    def __init__(self, data: dict):
        self.__data = data

    def load(self, fact: IConfigFactory, conf: IConfigSection) -> dict:
        return self.__data



from abc import ABC, abstractmethod
from typing import Union


class IConfigSection(ABC):
    @abstractmethod
    def get(self, key: str) -> any:
        raise NotImplemented()

    @abstractmethod
    def get_section(self, key: str) -> Union["IConfigSection", None]:
        raise NotImplemented()

    @abstractmethod
    def has(self, key: str) -> bool:
        raise NotImplemented()


class IConfigSource(ABC):
    @abstractmethod
    def load(self, fact: "IConfigFactory", conf: IConfigSection) -> dict:
        raise NotImplemented()


class IConfigFactory(ABC):
    @abstractmethod
    def add_source(self, source: IConfigSource) -> None:
        raise NotImplemented()

    @abstractmethod
    def build(self) -> IConfigSection:
        raise NotImplemented()

    @abstractmethod
    def add_configuration(self, conf: IConfigSection):
        raise NotImplemented()



from typing import List

from dnry.config.helpers import explode
from dnry.config.types import IConfigSource, IConfigFactory, IConfigSection
from argparse import ArgumentParser


class ArgumentSource(IConfigSource):
    """Read configuration values from an argparse.ArgumentParser object."""
    def __init__(self, argument_parser: ArgumentParser, argv: List[str] = None):
        self.__argv = argv
        self.__ap = argument_parser

    def load(self, fact: IConfigFactory, conf: IConfigSection) -> dict:
        opts = self.__ap.parse_args(self.__argv)
        return explode(vars(opts))



from typing import Callable, Union

from dnry.config.types import IConfigSource, IConfigFactory, IConfigSection


class DelegateSource(IConfigSource):
    def __init__(self, delegate: Callable[[IConfigFactory], Union[dict, None]]):
        self.__delegate = delegate

    def load(self, fact: IConfigFactory, conf: IConfigSection) -> dict:
        return self.__delegate(fact, conf) or {}



import os

from dnry.config.helpers import explode
from dnry.config.types import IConfigSource, IConfigFactory, IConfigSection


class EnvironmentSource(IConfigSource):
    """Read configuration values from Environment Variables.

    Reads all environment variables available to the process
    that starts with the optional prefix.  Keys are split on '_'
    and used as section separators.

    Example:
        /* ENV_KEY="hello" */
        hello = conf.get("ENV:KEY")
    """
    def __init__(self, prefix: str = None):
        self.__prefix = prefix or ""

    def load(self, fact: IConfigFactory, conf: IConfigSection) -> dict:
        val = {
            key[len(self.__prefix):].lstrip('_'): value
            for key, value
            in os.environ.items()
            if key.startswith(self.__prefix)}
        return explode(val, "_")




from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.readlines()

with open('README.md', "r") as f:
    long_description = f.read()

setup(
    name='dnry_config',
    version='0.1.0',
    author='Ian Laird',
    author_email='irlaird@gmail.com',
    url='https://github.com/en0/dnry-config',
    keywords=['config', 'configuration', 'dnry'],
    description='Multi-source config library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    packages=[
        'dnry.config',
        'dnry.config.arg',
        'dnry.config.delegate',
        'dnry.config.environ',
        'dnry.config.in_memory',
        'dnry.config.yaml',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=requirements,
)


from typing import Union

from dnry.config.types import IConfigSection


class ConfigSection(IConfigSection):
    """Basic configuration section implementation.

    The basic implementation provides access to nested key values
    by separating namespace using a colon ':'.

    For example, given the following logical configuration tree:
        { "parent": { "child": "value" } }

    You could access "value" by using
        configuration.get("parent:child")
    """
    def __init__(self, root: dict):
        self.__dict__ = root

    def has(self, key: str) -> bool:
        root = self.__dict__
        for k in key.split(":"):
            if k in root:
                root = root[k]
            else:
                return False
        return True

    def get(self, key: str) -> any:
        root = self.__dict__
        for k in key.split(":"):
            if k in root:
                root = root[k]
            else:
                return None
        return root

    def get_section(self, key: str) -> Union["IConfigSection", None]:
        val = self.get(key)
        if isinstance(val, dict):
            return ConfigSection(val)
        return None



import yaml
from typing import Union, List
from os.path import exists
from dnry.config.types import IConfigSource, IConfigFactory, IConfigSection


class YamlSource(IConfigSource):
    """Read configuration values from a yaml file."""
    def __init__(self, path: Union[str, List[str]], required: bool = True, **kwargs):
        self.__paths = path if isinstance(path, list) else [path]
        self.__required = required
        self.__loader = kwargs.get("loader", yaml.SafeLoader)

    def load(self, fact: IConfigFactory, conf: IConfigSection) -> dict:
        path = self.__get_first_existing_path()
        if path is None and self.__required:
            paths = ",".join(self.__paths)
            raise RuntimeError(f"Configuration Error: None of these paths could be found: {paths}")

        with open(path, 'r') as f:
            return yaml.load(f, Loader=self.__loader)

    def __get_first_existing_path(self) -> Union[str, None]:
        try:
            return next(p for p in self.__paths if exists(p))
        except StopIteration:
            return None



from typing import Dict, Union


def merge(a: dict, b: dict) -> dict:
    """Deep merge of 2 dict objects.

    Take the values in dict "b" and merge them
    into dict "a". If a key exists in both inputs
    the value from "b" will be used.
    """
    queue = list()
    ret = dict()
    queue.append((ret, a, b))
    while len(queue) > 0:
        root, left, right = queue.pop()
        for rk, rv in right.items():
            if rk not in left:
                root[rk] = rv
            elif isinstance(rv, list) and isinstance(left[rk], list):
                root[rk] = left[rk] + rv
            elif isinstance(rv, dict) and isinstance(left[rk], dict):
                root[rk] = {}
                queue.append((root[rk], left[rk], rv))
            else:
                root[rk] = rv
        for lk, lv in left.items():
            if lk not in right:
                root[lk] = lv
    return ret


def explode(items: Dict[str, Union[str, int, bool, list]], delimiter: str = ':') -> dict:
    """Explode a flat map into a nested map.

    Explode a flat map such as { "a:b:c", "val" } into
    a nested map such as { "a": { "b": { "c": "val" } } }
    """
    val = dict()
    for k, v in items.items():
        r = val
        parts = k.split(delimiter)
        for i in range(len(parts)):
            k_part = parts[i]
            if i == len(parts) - 1:
                if v is not None:
                    r[k_part] = v
            elif k_part not in r:
                r[k_part] = {}
                r = r[k_part]
            elif isinstance(r[k_part], dict):
                r = r[k_part]
    return val



import unittest

from dnry.config import ConfigFactory
from dnry.config.in_memory import InMemorySource


class TestFactory(unittest.TestCase):
    def test_single_key(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": 1}))
        conf = fact.build()
        val = conf.get("a")
        self.assertEqual(1, val)

    def test_multi_key(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": 1, "b": 2}))
        conf = fact.build()
        val = conf.get("a")
        self.assertEqual(1, val)

    def test_merge(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": 1}))
        fact.add_source(InMemorySource({"b": 2}))
        conf = fact.build()
        self.assertEqual(1, conf.get("a"))
        self.assertEqual(2, conf.get("b"))

    def test_merge_duplicate_key(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": 1}))
        fact.add_source(InMemorySource({"a": 2}))
        conf = fact.build()
        self.assertEqual(2, conf.get("a"))

    def test_merge_deep1(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": {"b": 1}}))
        fact.add_source(InMemorySource({"a": {"c": 2}}))
        conf = fact.build()
        self.assertEqual(1, conf.get("a:b"))
        self.assertEqual(2, conf.get("a:c"))

    def test_merge_deep2(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": {"b": 1, "c": 3}}))
        fact.add_source(InMemorySource({"a": {"c": 2}}))
        conf = fact.build()
        self.assertEqual(1, conf.get("a:b"))
        self.assertEqual(2, conf.get("a:c"))

    def test_merge_deep3(self):
        fact = ConfigFactory()
        fact.add_source(InMemorySource({"a": {"b": [1], "c": 3}}))
        fact.add_source(InMemorySource({"a": {"c": 2, "b": [2]}}))
        conf = fact.build()
        self.assertEqual(1, conf.get("a:b")[0])
        self.assertEqual(2, conf.get("a:b")[1])
